/* Do not edit - built by Makefile */
#ifndef FXLIB_H_INCLUDED
#define FXLIB_H_INCLUDED
#include <assert.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include "intvec.h"
/*
** $Id: fx_types.h,v 1.9 1995/10/17 05:26:10 davidj Exp $
**
** SPERT fixed point low level library - system-dependent definitions
** davidj - Thu Jan 27 16:14:46 1994
*/

#ifndef FXSYSTEM_H_INCLUDED
#define FXSYSTEM_H_INCLUDED


/* Some preprocessor symbols that should be in something like fx_config.h

FX_USE_SCALBN - indicates scalb() takes a double as the second argument
                and that scalbn should be used for integer second argumen
FX_HAVE_EXP2 - indicates exp2() exists in the standard math library
FX_HAVE_LOG2 - indicates log2() exists in the standard math library

*/


#if (defined(__sun__)||defined(sun)) && !defined(__svr4__)
#define FX_ARCH_SUN4 1
#define FX_MACH_SPARC 1
#define FX_USE_SCALBN
#define FX_HAVE_EXP2
#define FX_HAVE_LOG2
#elif defined(__torrent__)
#define FX_ARCH_SPERT0 1
#define FX_MACH_TORRENT 1
#define FX_MACH_T0 1
#define FX_MACH_MIPS 1
#else defined(__sgi__)
#define FX_ARCH_IRIS 1
#define FX_MACH_MIPS 1
#endif

/* The later version of gcc for SPERT uses newlib, which has scalbn and log2*/
#if defined(__torrent__) && (__GNUC_MINOR__>=7 || __GNUC__>2)
#define FX_USE_SCALBN
#define FX_HAVE_LOG2
#endif


#if defined(FX_MACH_SPARC) || defined(FX_MACH_MIPS) || defined(FX_MACH_TORRENT)

/* Some types we need */

typedef char fxint8;
typedef short fxint16;
typedef unsigned short fxuint16;
typedef int fxint32;
typedef unsigned int fxuint32;
typedef int fxbool;

typedef int fxexp_t;
typedef size_t fxsize_t;
typedef unsigned long fxtimer_t;

#define FXINT16_MAX (0x7fff)
#define FXINT16_MIN (-0x8000)
#define FXINT8_MAX (0x7f)
#define FXINT8_MIN (-0x80)
#define FXINT32_MAX (0x7fffffff)
#define FXINT32_MIN (-0x7fffffff-1)

#define FXUINT16_MAX (0xffffu)
#define FXUINT8_MAX (0xffu)
#define FXUINT32_MAX (0xffffffffu)

#define FXSIZE_MAX INT_MAX

/* Although exponents are normally 32 bits, we limit usable exponents to 16
   bits and so we can rely on the rest of the range to prevent overflows */

#define FXEXP_MAX FXINT16_MAX
#define FXEXP_MIN FXINT16_MIN

#endif

#if defined(FX_ARCH_SPERT0)
void* tormalloc(fxsize_t size);
void torfree(void *ptr);
#endif

#define FXOK (0)
#define FXERROR (-1)
#define FXTRUE (1)
#define FXFALSE (0)

/* This is the size of "gangs" on the T0 processor - some portable routines
   may require this value to get numerically identical results */

#define FX_GANG_SIZE (32)
#define FX_GANG_LOG2 (5)
#define FX_GANG_MASK (FX_GANG_SIZE-1)

#endif /* #ifndef FXSYSTEM_H_INCLUDED */

/*
** $Id: fx_excep.h,v 1.3 1994/12/15 01:17:13 davidj Exp $
*/


#ifndef FX_EXCEP_H_INCLUDED
#define FX_EXCEP_H_INCLUDED 

typedef void fxexcep_handler_type(void);

void fxexcep_init(fxexcep_handler_type users_unimp_handler,
		  fxexcep_handler_type users_exprange_handler);

void fxexcep_expunimp(void);
void fxexcep_exprange(void);
void fxexcep_expunimp_handler_default(void);
void fxexcep_exprange_handler_default(void);


#endif /* #ifndef FX_EXCEP_H_INCLUDED */
/*
** $Id: fx_int64.h,v 1.10 1994/12/13 00:59:33 davidj Exp $
**
** supposedly portable 64 bit arithmetic library
** DJ - Mon Feb 14 11:18:57 1994
*/

#ifndef FXINT64_H_INCLUDED
#define FXINT64_H_INCLUDED


typedef long long fxint64;
typedef unsigned long long fxuint64;

fxint64 conv_i32_i64(fxint32 a);
fxint32 conv_i64_i32(fxint64 a);
fxint16 conv_i64_i16(fxint64 a);
fxint32 high32_i64_i32(fxint64 a);
fxint64 sra_i64i_i64(fxint64 a, int b);
fxint64 sll_i64i_i64(fxint64 a, int b);
fxint64 neg_i64_i64(fxint64 a);
fxint64 add_i64i64_i64(fxint64 a, fxint64 b);
fxint64 sub_i64i64_i64(fxint64 a, fxint64 b);
fxbool setgt_i64i64_b(fxint64 a, fxint64 b);
fxbool setlt_i64i64_b(fxint64 a, fxint64 b);
fxbool seteq_i64i64_b(fxint64 a, fxint64 b);
fxbool iszero_i64_b(fxint64 a);
fxint64 zero_i64(void);
fxint64 one_i64(void);
fxint64 minusone_i64(void);
fxint64 maxint_i64(void);
fxint64 minint_i64(void);
fxint64 make_i32i32_i64(fxint32 hi, fxint32 lo);
fxint64 and_i64i64_i64(fxint64 a, fxint64 b);
fxint64 or_i64i64_i64(fxint64 a, fxint64 b);


#endif
/*
** $Id: fx_emu.h,v 1.6 1994/12/15 01:10:56 davidj Exp $
**
** Fixed point Torrent instruction emulation routines
** DJ - Fri Jul 29 16:25:16 1994
*/

#ifndef FXEMU_H_INCLUDED
#define FXEMU_H_INCLUDED

extern int fxemu_sat;

fxbool isoveradd_i32i32i32_b(fxint32 res, fxint32 a, fxint32 b);
fxbool isoversub_i32i32i32_b(fxint32 a_minus_b, fxint32 a, fxint32 b);

fxint8  emusat_i32_i8(fxint32 val);
fxint16 emusat_i32_i16(fxint32 val);
fxint32 emusadd_i32i32_i32(fxint32 a, fxint32 b);
fxint32 emussub_i32i32_i32(fxint32 a, fxint32 b);

fxint32 fxadd_i32i32_i32(int shift, fxint32 a, fxint32 b);
fxint16 fxadd_i32i32_i16(int shift, fxint32 a, fxint32 b);
fxint8  fxadd_i32i32_i8(int shift, fxint32 a, fxint32 b);
fxint32 fxmul_i16i16_i32(int rshift, fxint16 a, fxint16 b);
fxint16 fxmul_i16i16_i16(int rshift, fxint16 a, fxint16 b);
fxint8  fxmul_i16i16_i8(int rshift, fxint16 a, fxint16 b);


#endif /* #ifndef FXEMU_H_INCLUDED */
/*
** $Id: fx_ivec.h,v 1.24 1996/01/05 02:12:10 davidj Exp $
*/

#ifndef FX_IVEC_H_INCLUDED
#define FX_IVEC_H_INCLUDED


/* Scalar unuary */

fxint16 peg_i16_i16(fxint16 val);
fxint16 peg_i32_i16(fxint32 val);
fxint32 peg_i32_i32(fxint32 val);
fxint16 peg_i64_i16(fxint64 val);
fxint32 peg_i64_i32(fxint64 val);

fxint16 sat_i32_i16(fxint32 val);
fxint8 sat_i32_i8(fxint32 val);

fxbool sat_i32_i16s(fxint32 val, fxint16* res);
fxbool sat_i64_i16s(fxint64 val, fxint16* res);
fxbool sat_i64_i32s(fxint64 val, fxint32* res);

fxint32 log2_floor_i32_i32(fxint32 val);
fxint32 log2_ceil_i32_i32(fxint32 val);


/* Scalar binary */

fxbool satadd_i32i32_i32s(fxint32 a, fxint32 b, fxint32* res);
fxbool satsub_i32i32_i32s(fxint32 a, fxint32 b, fxint32* res);

fxint32 mul_i16i16_i32(fxint16 a, fxint16 b);

fxint16 satshl_i16i_i16(fxint16 val, int lshift);
fxint16 satshl_i32i_i16(fxint32 val, int lshift);
fxint32 satshl_i32i_i32(fxint32 val, int lshift);

fxbool satshl_i32i_i32s(fxint32 val, int lshift, fxint32* res);
fxbool satshl_i64i_i64s(fxint64 val, int lshift, fxint64* res);

fxint32 srl_i32i_i32(fxint32 val, int rshift);
fxint32 sra_i32i_i32(fxint32 val, int rshift);

fxint32 rndsrl_i32i_i32(fxint32 val, int rshift);

/*** Now in intvec
fxint32 satadd_i32i32_i32(fxint32 a, fxint32 b);
fxint32 satsub_i32i32_i32(fxint32 a, fxint32 b);
***/

/* Vector unary */

void peg_vi16_vi16(fxsize_t len, const fxint16* in, fxint16* res);
void peg_vi32_vi32(fxsize_t len, const fxint32* in, fxint32* res);
void peg_vi16_vi32(fxsize_t len, const fxint16* in, fxint32* res);

/* Vector-scalar binary */

/**** Now in intvec
void add_vi32i32_vi32(fxsize_t len, const fxint32* a, fxint32 b, fxint32* res);
void sub_vi32i32_vi32(fxsize_t len, const fxint32* a, fxint32 b, fxint32* res);
void mul_vi32i32_vi32(fxsize_t len, const fxint32* a, fxint32 b, fxint32* res);
void satadd_vi32i32_vi32(fxsize_t len, const fxint32* a, const fxint32 b,
			 fxint32* res);
void satsub_vi32i32_vi32(fxsize_t len, const fxint32* a, const fxint32 b,
			 fxint32* res);
***/


/* Vector-vector binary */
/**** Now in intvec
void add_vi32vi32_vi32(fxsize_t len, const fxint32* a, const fxint32* b,
		       fxint32* res);
void sub_vi32vi32_vi32(fxsize_t len, const fxint32* a, const fxint32* b,
		       fxint32* res);
void mul_vi32vi32_vi32(fxsize_t len, const fxint32* a, const fxint32* b,
		       fxint32* res);
void satadd_vi32vi32_vi32(fxsize_t len, const fxint32* a, const fxint32* b,
			  fxint32* res);
void satsub_vi32vi32_vi32(fxsize_t len, const fxint32* a, const fxint32* b,
			  fxint32* res);
***/

void satshl_vi16i_vi32(fxsize_t len, const fxint16* in, int lshift,
		       fxint32* res);

/* Redcutions */

fxint16 absxmax_vi16_i16(fxsize_t len, const fxint16* in);

/**** Now in intvec
fxint32 sum_vi32_i32(fxsize_t len, const fxint32* in);
***/
fxint32 absxmax_vi32_i32(fxsize_t len, const fxint32* in);


#endif /* #ifndef FX_IVEC_H_INCLUDED */
/*
** $Id: fx_xvec.h,v 1.19 1996/10/24 23:09:55 davidj Exp $
*/

#ifndef FX_XVEC_H_INCLUDED
#define FX_XVEC_H_INCLUDED


/* Scalar fixed point operations */

fxint16 srx_i16i_i16(fxint16 val, int rshift);
fxint32 srx_i32i_i32(fxint32 val, int rshift);
fxint64 srx_i64i_i64(fxint64 val, int rshift);
fxint32 slx_i32i_i32(fxint32 val, int lshift);
fxint16 mul_x16x16_x16(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                       fxint16 x, fxint16 y);
fxint16 add_x16x16_x16(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                       fxint16 x, fxint16 y);
fxint16 sub_x16x16_x16(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                       fxint16 x, fxint16 y);
fxint32 mul_x16x16_x32(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                       fxint16 x, fxint16 y);
fxint32 add_x32x32_x32(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                       fxint32 x, fxint32 y);
fxint32 sub_x32x32_x32(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                       fxint32 x, fxint32 y);

/*
** Scalar helper functions
*/

fxint32 _add_x32x32_x32__ls(fxexp_t shift, fxint32 x, fxint32 y);
fxint32 _sub_x32x32_x32__ls(fxexp_t shift, fxint32 x, fxint32 y);

/* Vector operations */

void acc_vx32_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		   const fxint32* in, fxint32* res);
void dsigmoid_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
			const fxint16* in, fxint16* res);

/* Vector-scalar operations */

void srx_vi16i_vi16(fxsize_t len, const fxint16* in, int rshift, 
		    fxint16* res);
void srx_vi16i_vi32(fxsize_t len, const fxint16* in, int rshift, 
		    fxint32* res);
void slx_vi16i_vi32(fxsize_t len, const fxint16* in, int rshift, 
		    fxint32* res);
void slx_vi32i_vi32(fxsize_t len, const fxint32* in, int lshift,
		    fxint32* res);
void add_vx16x16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
		      fxexp_t r_exp, const fxint16* x, fxint16 y,
		      fxint16* r);
void mul_vx16x16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
		      fxexp_t r_exp, const fxint16* x, fxint16 y,
		      fxint16* r);
void mul_vx16x16_vx32(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
		      fxexp_t r_exp, const fxint16* x, fxint16 y,
		      fxint32* r);
fxbool mulacc_vx16x16_vx16s(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
			    fxexp_t r_exp, const fxint16* x, fxint16 y,
			    fxint16* r);
void mulacc_vx16x16_vx32(fxsize_t len, fxexp_t a_exp, fxexp_t b_exp,
			 fxexp_t res_exp, const fxint16* a, fxint16 b,
			 fxint32* res);

/* Vector-scalar helper functions */

fxbool mul_srx_acc_vi16i16_i_vi16s(fxsize_t len, const fxint16* a,
				   fxint16 b, fxexp_t rshift, fxint16* res);

/*
** Scalar-vector operations
*/

void sub_x16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
		      fxexp_t r_exp, fxint16 x, const fxint16* y,
		      fxint16* r);
/*
void add_x32vx32_vx32(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
		      fxexp_t r_exp, fxint32 x, const fxint32* y,
		      fxint32* r);
void sub_x32vx32_vx32(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
		      fxexp_t r_exp, fxint32 x, const fxint32* y,
		      fxint32* r);
*/

/* Scalar-vector helper functions */

void _add_x16vx16_vx16__lr(fxsize_t size, fxexp_t lshift, fxexp_t rshift,
			fxint16 x, const fxint16* y, fxint16* r);
void _sub_x16vx16_vx16__lr(fxsize_t size, fxexp_t lshift, fxexp_t rshift,
			fxint16 x, const fxint16* y, fxint16* r);
void _mul_x16vx16_vx32__r(fxsize_t size, fxexp_t rshift, fxint16 a,
		       const fxint16* b, fxint32* res);
void _mul_x16vx16_vx32__l1(fxsize_t size, fxint16 a, const fxint16* b,
			fxint32* res);
/*
** Vector-vector element wise fixed point operations
*/

void mul_vx16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp, 
                       fxexp_t r_exp, const fxint16* x, const fxint16* y,
                       fxint16* r);
void add_vx16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
                       fxexp_t r_exp, const fxint16* x, const fxint16* y,
                       fxint16* r);
void sub_vx16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
                       fxexp_t r_exp, const fxint16* x, const fxint16* y,
                       fxint16* r);
void mul_vx16vx16_vx32(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
                       fxexp_t r_exp, const fxint16* x, const fxint16* y,
                       fxint32* r);
void add_vx32vx32_vx32(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
                       fxexp_t r_exp, const fxint32* x, const fxint32* y,
                       fxint32* r);
void sub_vx32vx32_vx32(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
                       fxexp_t r_exp, const fxint32* x, const fxint32* y,
                       fxint32* r);

/*
** The functions that actually do the work for the vector functions
*/

void mul_srx_vi16i16_i_vi16(fxsize_t size,  const fxint16* a, fxint16 b,
			    fxexp_t rshift, fxint16* res);

void _add_vx16vx16_vx16__lr(fxsize_t size, fxexp_t lshift, fxexp_t rshift,
			    const fxint16* a, const fxint16* b, fxint16* res);
void _sub_vx16vx16_vx16__lr(fxsize_t size, fxexp_t lshift, fxexp_t rshift,
			    const fxint16* a, const fxint16* b, fxint16* res);
void _mul_vx16vx16_vx16__r(fxsize_t size,  fxexp_t rshift, const fxint16* a,
			   const fxint16* b, fxint16* res);
void _add_vx32vx32_vx32__ls(fxsize_t size, fxexp_t shift,
			    const fxint32* a, const fxint32* b, fxint32* res);
void _add_vx32vx32_vx32__zz(fxsize_t size, const fxint32* a, const fxint32* b,
			    fxint32* res);
void _sub_vx32vx32_vx32__ls(fxsize_t size, fxexp_t shift,
			    const fxint32* a, const fxint32* b, fxint32* res);
void _sub_vx32vx32_vx32__zz(fxsize_t size, const fxint32* a, const fxint32* b,
			    fxint32* res);
void _mul_vx16vx16_vx32__r(fxsize_t size, fxexp_t shift, const fxint16* a,
			   const fxint16* b, fxint32* res);
void _mul_vx16vx16_vx32__l1(fxsize_t size, const fxint16* a, const fxint16* b,
			    fxint32* res);

/*
** Reducing fixed point operations
*/

/* Dot product */
fxbool mulsum_vx16vx16_x32t(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp,
			    fxexp_t r_exp, const fxint16* x, const fxint16* y,
			    fxint32* res);

/*
** Reducing fixed point operations worker functions
*/
fxbool mul_srx_sum_vi16vi16_i_i32t(fxsize_t size, const fxint16* a,
				   const fxint16* b, fxexp_t rshift,
				   fxint32* res);

/*
** Matrix fixed point operations
*/

fxbool mul_vx16mx16_vx32t(fxsize_t height, fxsize_t width, fxexp_t vec_exp,
			  fxexp_t mat_exp, fxexp_t res_exp, const fxint16* vec,
			  const fxint16* mat, fxint32* res);
fxbool mul_mx16vx16_vx32t(fxsize_t height, fxsize_t width, fxexp_t mat_exp,
			  fxexp_t vec_exp, fxexp_t res_exp,
			  const fxint16* mat, const fxint16* vec,fxint32* res);
fxbool mulacc_vx16vx16_mx16s(fxsize_t height, fxsize_t width,
			     fxexp_t a_exp, fxexp_t b_exp, fxexp_t res_exp,
			     const fxint16* a, const fxint16* b, fxint16* res);
fxbool
mulacc_vx16vx16_mhx32s(fxsize_t height, fxsize_t width, fxsize_t stride,
		       fxexp_t a_exp, fxexp_t b_exp, fxexp_t res_exp,
		       const fxint16* a, const fxint16* b, fxint16* resh,
		       fxint16* resl);


/*
** Matrix functions that actually do the work
*/

fxbool mul_srx_vi16mi16_i_vi32t(fxsize_t height, fxsize_t width,
				const fxint16* vec, const fxint16* mat,
				fxexp_t rshift, fxint32* res);
fxbool mul_srx_mi16vi16_i_vi32t(fxsize_t height, fxsize_t width,
				const fxint16* mat, const fxint16* vec,
				fxexp_t rshift, fxint32* res);
fxbool mul_srx_acc_vi16vi16_i_mi16s(fxsize_t height, fxsize_t width,
				    const fxint16* a, const fxint16* b,
				    fxexp_t rshift, fxint16* res);
fxbool
mul_srx_acc_vi16vi16_i_mhi32s(fxsize_t height, fxsize_t width, fxsize_t stride,
			      const fxint16* a, const fxint16* b,
			      fxexp_t rshift, fxint16* resh, fxint16* resl);

#endif /* #ifndef FX_XVEC_H_INCLUDED */
/*
** $Id: fx_conv.h,v 1.16 1995/12/22 06:24:36 davidj Exp $
*/

#ifndef FXCONV_H_INCLUDED
#define FXCONV_H_INCLUDED


fxint16 conv_x16_x16(fxexp_t in_exp, fxexp_t res_exp, fxint16 in);
fxint32 conv_x16_x32(fxexp_t in_exp, fxexp_t res_exp, fxint16 in);
fxint16 conv_x32_x16(fxexp_t in_exp, fxexp_t res_exp, fxint32 in);
fxint32 conv_x32_x32(fxexp_t in_exp, fxexp_t res_exp, fxint32 in);

fxbool conv_f_x16s(fxexp_t res_exp, float val, fxint16* res);
fxbool conv_f_x32s(fxexp_t res_exp, float val, fxint32* res);
fxint16 conv_f_x16(fxexp_t res_exp, float val);
fxint32 conv_f_x32(fxexp_t res_exp, float val);
fxbool conv_d_x16s(fxexp_t res_exp, double val, fxint16* res);
fxint16 conv_d_x16(fxexp_t res_exp, double val);
double conv_x16_d(fxexp_t in_exp, fxint16 val);
float conv_x16_f(fxexp_t in_exp, fxint16 val);
double conv_x32_d(fxexp_t in_exp, fxint32 val);
float conv_x32_f(fxexp_t in_exp, fxint32 fval);

fxint16 ceil_d_x16(fxexp_t res_exp,  double val);
fxint16 floor_d_x16(fxexp_t res_exp, double val);
fxint32 ceil_d_x32(fxexp_t res_exp,  double val);
fxint32 floor_d_x32(fxexp_t res_exp, double val);

/* This would not be necessary if the World could agree on what scalb()
   does */
double mulpow2_di_d(double in, int n);

void conv_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		    const fxint16* in, fxint16* res);
void conv_vx16_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		    const fxint16* in, fxint32* res);
void conv_vx32_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		    const fxint32* in, fxint16* res);
void conv_vx32_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		    const fxint32* in, fxint32* res);

fxbool conv_vf_vx16s(fxsize_t len, fxexp_t res_exp, const float* in,
		     fxint16* res);
fxbool conv_vf_vx32s(fxsize_t len, fxexp_t res_exp, const float* in,
		     fxint32* res);
void conv_vx16_vf(fxsize_t len, fxexp_t exp, const fxint16* in, float* res);
void conv_vx32_vf(fxsize_t len, fxexp_t exp, const fxint32* in, float* res);

void conv_vi16_vi32(fxsize_t len, const fxint16* in, fxint32* res);

void _conv_vx16_vx16__l(fxsize_t len, fxexp_t lshift, const fxint16* in,
			fxint16* res);
void _conv_vx32_vx16__l(fxsize_t len, fxexp_t lshift, const fxint32* in,
			fxint16* res);
void _conv_vx32_vx16__r(fxsize_t len, fxexp_t rshift, const fxint32* in,
			fxint16* res);
void _conv_vx32_vx32__l(fxsize_t len, fxexp_t lshift, const fxint32* in,
			fxint32* res);
void _conv_vx32_vx32__l32(fxsize_t len, const fxint32* in,
			  fxint32* res);
void _conv_vx32_vx32__r(fxsize_t len, fxexp_t rshift, const fxint32* in,
			fxint32* res);


#endif /* #ifndef FXCONV_H_INCLUDED */
/*
** $Id: fx_table.h,v 1.12 1995/10/11 02:54:40 davidj Exp $
**
** davidj - Thu Aug 25 16:00:58 1994
*/

#ifndef FXTABLE_H_INCLUDED
#define FXTABLE_H_INCLUDED

/*
pltables can be used to store arbitrary functions of one variable.  They
support a fixed absolute accuracy over a fixed domain:  The function argument
and result are both assumed to be 16-bit fixed-point numbers with an exponent
of zero.  (Hence, both input and output are over the range [-1,1).)  For
function results out of range, saturation of the result occurs automatically.
    Functions are approximated using a piecewise linear technique.  Each linear
piece is represented by a constant term (base) and a linear term (slope).  To
save space, the slope of every linear piece must be within the range [-64,64).
The total space occupied by a pltable is always exactly 8 kB.
    For ``well-behaved'' functions, the worst case error of the pltable
approximation should be under 1 ulp (unit in the last place), and is usually
around 0.8 ulps.  If the slope (first derivative) of the desired function ever
exceeds +/-64, the slope of some of the linear pieces can saturate, and this
can result in arbitrarily large error.  Extra error can also arise if the
curvature (second derivative) of the function exceeds what the table is capable
of resolving.  Curvature of much more than +/-128 is likely to cause trouble.
The square root function is inherently problematic, for instance, because the
slope and curvature are both infinite at 0.
    A pltable divides a function into 2048 pieces, with each piece approximated
by a separate line segment.  Each table entry corresponds to one linear piece.
A table entry is 32 bits in size:  The upper 19 bits hold the base of the
linear piece, and the lower 13 bits hold the slope.  The base can be considered
a 19-bit fixed-point value with an exponent of 1.  The slope is a 13-bit fixed-
point value with an exponent of 6.
*/


#define FX_PLTABLE_SIZE 2048
typedef fxint32 fx_pltable_t;

/* A routine to fill a pltable given a function which the table must
   approximate */
void fill_pltable_func_d_d(fx_pltable_t* tbl, fxexp_t domain_exp,
			   fxexp_t range_exp, double (*func)(double x) );
/* A routine to create a pltable and write it out as a C source file */
int create_pltable(const char* name, fxexp_t domain_exp, fxexp_t range_exp,
		   double (*func)(double x),
		   double start, double end, FILE* outStream);
/* Display the error between the function stored in a pltable and the double
   precision version of that function */
fxbool showerr_pltable_func_d_d(const fx_pltable_t* tbl,
				fxexp_t domainExp, fxexp_t rangeExp,
				double (*func)(double x),
				double startA, double endA, FILE* outStream);
/* Lookup a specific value in a pltable */
fxint16 lkup_pltable_i16_i16(const fx_pltable_t* tbl, fxint16 a);
fxbool lkup_pltable_i16_i16s(const fx_pltable_t* tbl, fxint16 in,
			     fxint16* res);
/* Lookup a vector in a pltable */
fxbool lkup_pltable_vi16_vi16s(fxsize_t len, const fx_pltable_t* tbl,
			       const fxint16* in, fxint16* res);
#define lkup_pltable_vi16_vi16(a, b, c, d) \
        lkup_pltable_vi16_vi16s((a), (b), (c), (d))

/* gtables are based on pltables but use friendlier wrapper routines.  They
   handle general fixed point numbers as input and output, as opposed to fixed
   point numbers with specific implied exponents that pltables use.  It is
   intended that gtables will be use by library users to generate their own
   tables, whereas pltables will typically be used by other routines in the
   library.  */

/* Here is the structure used to store a gtable */

typedef struct fx_gtable_s
{
    fxexp_t domain_exp;		/* Exponent of number spanning domain
				   of table */
    fxexp_t range_exp;		/* Exponent of data in table */
    fx_pltable_t data[FX_PLTABLE_SIZE]; /* The actual data in the table */
} fx_gtable_t;

/* A routine to fill a gtable given a function which the table must
   approximate */
void fill_gtable_func_d_d(fx_gtable_t* tbl, double (*func)(double x));
/* A routine to create a gtable and write it out as a C source file */
int create_gtable(const char* name, fxexp_t domain_exp, fxexp_t range_exp,
		  double (*func)(double x),
		  double start, double end, FILE* outStream);
/* Display the error between the function stored in a gtable and the double
   precision version of that function */
fxbool showerr_gtable_func_d_d(const fx_gtable_t* tbl,double (*func)(double x),
			     double start, double end, FILE* outStream);
/* Lookup a specific value in a gtable */
fxint16 lkup_gtable_x16_x16(const fx_gtable_t* tbl, fxexp_t in_exp,
			    fxexp_t res_exp, fxint16 val);
/* Lookup up a fixed point vector in a gtable */
void lkup_gtable_vx16_vx16(fxsize_t len, const fx_gtable_t* tbl,
			   fxexp_t in_exp, fxexp_t res_exp,
			   const fxint16* in, fxint16* res);


/**************************
******** INPORTANT ********

pltables make utables redundant.
utables are not completed or tested, but have not been removed from the
library as they are almost finished and may yet be useful.
davidj - Tue Nov 29 13:08:19 1994

**************************/

/* utable16's are simple tables of 16 bit values, symmetric around 0
   and with powers of 2 length.  This structure holds all the
   information pertinent to a given utable16.

   Note that in a utable16, the input to the table is NOT rounded for lookup -
   it is simply truncated.  Each location in the table with an address
   correspoing to value x_n must hold an output appropriate for any value
   between x_n and x_(n+1), where x_(n+1) is the input value corresponding to
   the next table entry.
*/

typedef struct
{
    fxint16* data;		/* Point to the start of the table data */
    fxsize_t len;		/* Length of the table data (in halfs) */
    fxexp_t domain_exp;		/* Exponent of number spanning domain
				   of table */
    fxexp_t range_exp;		/* Exponent of data in table */

/* These are some values calculated at table initialization time to speed
   table lookup */

    fxint16* base;		/* Pointer to middle of table */
    fxexp_t in_rshift_bias;	/* Number to add to input exponent to
				   get input value right shift */
} fx_utable16_t;


/* Functions for generating, filling etc. tables */

void init_utable16(fx_utable16_t* table, fxint16* data,
		   fxsize_t len, fxexp_t domain_exp, fxexp_t range_exp);
void fill_utable16_func_d_d(fx_utable16_t* table, double (*func)(double x));

/* General table lookup functions */

fxint16 lkup_utable16_x16_x16(fx_utable16_t* table, fxexp_t in_exp,
			      fxexp_t res_exp, fxint16 in);
void lkup_utable16_vx16_vx16(fxsize_t len, fx_utable16_t* table,
			     fxexp_t in_exp, fxexp_t res_exp,
			     fxint16* in, fxint16* res);

/* Table lookup helper functions */

fxint16 _lkup_utable16_x16_x16__rr(fxint16* base, fxexp_t in_rshift,
				   fxexp_t out_rshift, fxint16 val);
void _lkup_utable16_vx16_vx16__rz(fxsize_t len, fxint16* base,
				  fxexp_t in_rshift,fxint16* in, fxint16* res);


#endif /* #ifndef FXTABLE_H_INCLUDED */
/*
** $Id: fx_func.h,v 1.14 1995/10/12 00:54:05 davidj Exp $
**
** davidj - Mon Oct 17 17:54:35 1994
*/

#ifndef FX_FUNC_H_INCLUDED
#define FX_FUNC_H_INCLUDED

extern fx_pltable_t pltable_logadd13[FX_PLTABLE_SIZE];
extern fx_pltable_t pltable_exp13[FX_PLTABLE_SIZE];
extern fx_pltable_t pltable_sigmoid_4_0[FX_PLTABLE_SIZE];
extern fx_pltable_t pltable_tanh_3_0[FX_PLTABLE_SIZE];

/* These are some functions based on the pltable and gtable tables */

/* sigmoid - 1/(1 + e^-x) */
fxint16 sigmoid_x16_x16(fxexp_t in_exp, fxexp_t res_exp, fxint16 in);
void sigmoid_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		       const fxint16* in, fxint16* res);
/* tanh */
fxint16 tanh_x16_x16(fxexp_t in_exp, fxexp_t res_exp, fxint16 in);
void tanh_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		       const fxint16* in, fxint16* res);

/* logadd - log_e(e^a + e^b) */
fxint32 logadd_x32x32_x32(fxexp_t a_exp, fxexp_t b_exp, fxexp_t res_exp,
			  fxint32 a, fxint32 b);
/* exponential */
fxbool exp_x32_x32s(fxexp_t in_exp, fxexp_t res_exp, fxint32 in, fxint32* res);
fxbool exp_vx32_vx32s(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		      const fxint32* in, fxint32* res);

/* The logsum function - a reducing logadd (a 'norm' in the log domain!) */
fxint32 logsum_vx32_x32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
			const fxint32* in);

/* The log of the softmax function: in[i] - logsum(in) */
void logsoftmax_vx32_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
			  const fxint32* in, fxint32* res);
/* The softmax function - exp(in[i]) / sum(exp[in[i]) */
void softmax_vx32_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		       const fxint32* in, fxint32* res);
void softmax_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		       const fxint16* in, fxint16* res);
void softmax_vx32_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		       const fxint32* in, fxint16* res);


/* This logadd13 function actually takes in fixed point numbers with an
   implied exponent of 13 */
fxint32 logadd13_i32i32_i32(fxint32 a, fxint32 b);

/* This logsum13 finds the sum of the logs of a vector of numbers - all
   values have an implied exponent of 13 */
fxint32 logsum13_vi32_i32(fxsize_t len, const fxint32* in);

fxbool exp13_i32_x32s(fxexp_t z_exp, fxint32 a, fxint32 *zPtr );
fxbool exp13_vi32_vx32s(fxsize_t len, fxexp_t in_exp, const fxint32* in,
			fxint32* res);


#endif FX_FUNC_H_INCLUDED

#ifdef FX_INLINED
#  define _FX_INLINED FX_INLINED
#  undef FX_INLINED
#endif
#ifdef __GNUC__
#  define FX_INLINED extern inline
#endif

FX_INLINED fxint32
fxadd_i32i32_i32(int shift, fxint32 a, fxint32 b)
{
    fxint32 res;		/* Result */
    fxint64 sum;		/* Sum of shifted a and b */
    fxint32 sticky;		/* Sticky bit */
    fxint32 temp;		/* Intermediate result */
    const fxuint32 MASK_LOW32 = 0xffffffffu;

    assert(shift>=0 && shift<32);
    if (shift)
        a = ((a<<shift) & MASK_LOW32) + (1 << (shift-1));
    sum = add_i64i64_i64(conv_i32_i64(a), conv_i32_i64(b));
    if (shift)
    {
        temp = conv_i64_i32(sra_i64i_i64(sum, shift));
        sticky = (conv_i64_i32(sum) << (32 - shift)) & MASK_LOW32;
        if (!sticky)
            temp &= 0xfffffffe;
        res = temp;
    }
    else
    {
        if (setgt_i64i64_b(sum, conv_i32_i64(FXINT32_MAX)))
        {
            res = FXINT32_MAX;
            fxemu_sat = 1;
        }
        else if (setlt_i64i64_b(sum, conv_i32_i64(FXINT32_MIN)))
        {
            res = FXINT32_MIN;
            fxemu_sat = 1;
        }
        else
        {
            res = conv_i64_i32(sum);
        }
    }
    return(res);
}

FX_INLINED fxint16
emusat_i32_i16(fxint32 val)
{
    if (val>FXINT16_MAX)
    {
	fxemu_sat = 1;
        return(FXINT16_MAX);
    }
    else if (val<FXINT16_MIN)
    {
	fxemu_sat = 1;
        return(FXINT16_MIN);
    }
    else
        return((fxint16) val);
}

FX_INLINED fxint16
fxadd_i32i32_i16(int shift, fxint32 a, fxint32 b)
{
    fxint32 temp;
    fxint16 res;

    temp = fxadd_i32i32_i32(shift, a, b);
    res = emusat_i32_i16(temp);
    return(res);
}

FX_INLINED fxint8
emusat_i32_i8(fxint32 val)
{
    if (val>FXINT8_MAX)
        return(FXINT8_MAX);
    else if (val<FXINT8_MIN)
        return(FXINT8_MIN);
    else
        return((fxint8) val);
}

FX_INLINED fxint8
fxadd_i32i32_i8(int shift, fxint32 a, fxint32 b)
{
    fxint32 temp;
    fxint8 res;

    temp = fxadd_i32i32_i32(shift, a, b);
    res = emusat_i32_i8(temp);
    return(res);
}

FX_INLINED fxint32
fxmul_i16i16_i32(int rshift, fxint16 a, fxint16 b)
{
    fxint32 res;		/* Result */
    fxint32 mult;		/* Multiplication result */
    fxint32 sticky;		/* Sticky bit */
    const fxuint32 MASK_LOW32 = 0xffffffffu;

    assert(rshift>=0 && rshift<32);
    mult = a * b;
    /* Only one case when we need 33 bit arithmetic - result in
       this case is 0 */
    if ((mult==0x40000000) && (rshift==31))
        res = 0;
    else
    {
        if (rshift)
        {
            mult += 1 << (rshift-1);
            res = mult >> rshift;
            sticky = (mult << (32 - rshift)) & MASK_LOW32;
            if (!sticky)
                res &= 0xfffffffe;
        }
        else
            res = mult;
    }
    return(res);
}

FX_INLINED fxint16
fxmul_i16i16_i16(int rshift, fxint16 a, fxint16 b)
{
    fxint32 temp;
    fxint16 res;

    temp = fxmul_i16i16_i32(rshift, a, b);
    res = emusat_i32_i16(temp);
    return(res);
}

FX_INLINED fxint8
fxmul_i16i16_i8(int rshift, fxint16 a, fxint16 b)
{
    fxint32 temp;
    fxint8 res;

    temp = fxmul_i16i16_i32(rshift, a, b);
    res = emusat_i32_i8(temp);
    return(res);
}

FX_INLINED fxbool
isoveradd_i32i32i32_b(fxint32 res, fxint32 a, fxint32 b)
{
    return (((a^b) >= 0) & ((a^res) < 0));
}

FX_INLINED fxint32
emusadd_i32i32_i32(fxint32 a, fxint32 b)
{
    fxint32 res;

    res = a + b;

    if (isoveradd_i32i32i32_b(res, a, b))
    {
	fxemu_sat = 1;
	if (res>=0)
	    res = FXINT32_MIN;
	else
	    res = FXINT32_MAX;
    }
    return(res);
}

FX_INLINED fxbool
isoversub_i32i32i32_b(fxint32 a_minus_b, fxint32 a, fxint32 b)
{
    return(((a^b) < 0) & ((a^a_minus_b) < 0));
}

FX_INLINED fxint32
emussub_i32i32_i32(fxint32 a, fxint32 b)
{
    fxint32 res;

    res = a - b;

    if (isoversub_i32i32i32_b(res, a, b))
    {
	fxemu_sat = 1;
	if (res>0)
	    res = FXINT32_MIN;
	else
	    res = FXINT32_MAX;
    }
    return(res);
}

FX_INLINED fxint32
log2_ceil_i32_i32(fxint32 val)
{
    fxint32 res;

    assert(val>0);

    res = 32 - countlz_i32_i32(val-1);
    return(res);
}

FX_INLINED fxint32
log2_floor_i32_i32(fxint32 val)
{
    fxint32 res;

    assert(val>0);

    res = 32 - countlz_i32_i32(val) - 1;
    return(res);
}

FX_INLINED fxint16
peg_i16_i16(fxint16 val)
{
    if (val>0)
        return(FXINT16_MAX);
    else if (val<0)
        return(FXINT16_MIN);
    else
        return(0);
}

FX_INLINED fxint16
peg_i32_i16(fxint32 val)
{
    if (val>0)
        return(FXINT16_MAX);
    else if (val<0)
        return(FXINT16_MIN);
    else
        return(0);
}

FX_INLINED fxint32
peg_i32_i32(fxint32 val)
{
    if (val>0)
        return(FXINT32_MAX);
    else if (val<0)
        return(FXINT32_MIN);
    else
        return(0);
}

FX_INLINED fxint16
peg_i64_i16(fxint64 val)
{
    if (setgt_i64i64_b(val, zero_i64()))
        return(FXINT16_MAX);
    else if (setlt_i64i64_b(val, zero_i64()))
        return(FXINT16_MIN);
    else
        return(0);
}

FX_INLINED fxint32
peg_i64_i32(fxint64 val)
{
    if (setgt_i64i64_b(val, zero_i64()))
        return(FXINT32_MAX);
    else if (setlt_i64i64_b(val, zero_i64()))
        return(FXINT32_MIN);
    else
        return(0);
}

FX_INLINED fxint16
sat_i32_i16(fxint32 val)
{
    if (val>FXINT16_MAX)
        return(FXINT16_MAX);
    else if (val<FXINT16_MIN)
        return(FXINT16_MIN);
    else
        return((fxint16) val);
}

FX_INLINED fxint8
sat_i32_i8(fxint32 val)
{
    if (val>FXINT8_MAX)
        return(FXINT8_MAX);
    else if (val<FXINT8_MIN)
        return(FXINT8_MIN);
    else
        return((fxint8) val);
}

FX_INLINED fxbool
sat_i32_i16s(fxint32 val, fxint16* res)
{
    if (val>FXINT16_MAX)
    {
	*res = FXINT16_MAX;
	return(1);
    }
    else if (val<FXINT16_MIN)
    {
	*res = FXINT16_MIN;
	return(1);
    }
    else
    {
	*res = (fxint16) val;
	return(0);
    }
}

FX_INLINED fxbool
sat_i64_i16s(fxint64 val, fxint16* res)
{
    if (setgt_i64i64_b(val, conv_i32_i64(FXINT16_MAX)))
    {
	*res = FXINT16_MAX;
        return(1);
    }
    else if (setlt_i64i64_b(val, conv_i32_i64(FXINT16_MIN)))
    {
        *res = FXINT16_MIN;
	return(1);
    }
    else
    {
        *res = conv_i64_i16(val);
	return(0);
    }
}

FX_INLINED fxbool
sat_i64_i32s(fxint64 val, fxint32* res)
{
    if (setgt_i64i64_b(val, conv_i32_i64(FXINT32_MAX)))
    {
        *res = FXINT32_MAX;
	return(1);
    }
    else if (setlt_i64i64_b(val, conv_i32_i64(FXINT32_MIN)))
    {
        *res = FXINT32_MIN;
	return(1);
    }
    else
    {
        *res = conv_i64_i32(val);
	return(0);
    }
}

FX_INLINED fxint32
sra_i32i_i32(fxint32 val, int rshift)
{
    return(val >> rshift);
}

FX_INLINED fxint32
srl_i32i_i32(fxint32 val, int rshift)
{
    return(((fxuint32) val) >> rshift);
}

FX_INLINED fxint32
mul_i16i16_i32(fxint16 a, fxint16 b)
{
    return((fxint32) a * (fxint32) b);
}

FX_INLINED fxint16
satshl_i16i_i16(fxint16 val, int lshift)
{
    fxint32 temp;
    fxint16 res;

    assert(((unsigned) lshift) < 16);

    temp = val;
    res = sat_i32_i16(temp << lshift);
    return(res);
}

FX_INLINED fxint32
satshl_i32i_i32(fxint32 val, int lshift)
{
    int sign;
    fxuint32 mask;
    fxint32 res;

    assert(((unsigned) lshift) < 32);
    sign = (val < 0);
    mask = 0xffffffff << (32 - lshift - 1);
    if (!sign)
    {
        if (((fxuint32) val & mask) != 0)
            res = FXINT32_MAX;
        else
            res = val << lshift;
    }
    else
    {
        if (((fxuint32) val & mask) != mask)
            res = FXINT32_MIN;
        else
            res = val << lshift;
    }
    return(res);
}

FX_INLINED fxint16
satshl_i32i_i16(fxint32 val, int lshift)
{
    fxint32 temp;
    fxint16 res;

    assert(((unsigned) lshift) < 16);
/*
** Note - to do a left shift (0 < shift <= 16), we need to clip first
** otherwise we will lose the sign bit.
*/
    temp = (fxint32) sat_i32_i16(val);
    res = sat_i32_i16(temp << lshift);
    return(res);
}

FX_INLINED fxbool
satshl_i32i_i32s(fxint32 val, int lshift, fxint32* res)
{
    int sign;
    fxuint32 mask;
    fxbool sat = 0;

    assert(((unsigned) lshift) < 32);
    sign = (val < 0);
    mask = 0xffffffff << (32 - lshift - 1);
    if (!sign)
    {
        if (((fxuint32) val & mask) != 0)
	{
            *res = FXINT32_MAX;
	    sat = 1;
	}
        else
            *res = val << lshift;
    }
    else
    {
        if (((fxuint32) val & mask) != mask)
	{
            *res = FXINT32_MIN;
	    sat = 1;
	}
        else
            *res = val << lshift;
    }
    return(sat);
}

FX_INLINED fxbool
satshl_i64i_i64s(fxint64 val, int lshift, fxint64* res)
{
    fxbool sat;
    fxint64 topbits;

    assert(((unsigned) lshift) < 64);
    if (lshift==0)
    {
	*res = val;
	sat = 0;
    }
    else
    {
	topbits = sra_i64i_i64(val, 64-lshift);
	if (setlt_i64i64_b(topbits, zero_i64()))
	{
	    if (seteq_i64i64_b(topbits, minusone_i64()))
	    {
		*res = sll_i64i_i64(val, lshift);
		sat = 0;
	    }
	    else
	    {
		*res = minint_i64();
		sat = 1;
	    }
	}
	else
	{
	    if (iszero_i64_b(topbits))
	    {
		*res = sll_i64i_i64(val, lshift);
		sat = 0;
	    }
	    else
	    {
		*res = maxint_i64();
		sat = 1;
	    }
	}
    }
    return(sat);
}

FX_INLINED fxint32
rndsrl_i32i_i32(fxint32 val, int rshift)
{
    fxint32 res;		/* Result */
    fxuint32 rem;		/* Bits shifted off right */

    assert(((unsigned) rshift) < 32);
    if (rshift==0)
	res = val;
    else
    {
	res = srl_i32i_i32(val, rshift);
	rem = val << (32-rshift);
	res += srl_i32i_i32((fxint32) rem, 31); /* If rem has MSB set, add 1 to res */
	if (rem==0x80000000)
	    res &= 0xfffffffe;
    }
    return(res);
}

FX_INLINED fxbool
satadd_i32i32_i32s(fxint32 a, fxint32 b, fxint32* res)
{
    fxint32 temp;
    fxbool sat = FXFALSE;

    temp = a + b;

    if (isoveradd_i32i32i32_b(temp, a, b))
    {
	sat = FXTRUE;
	if (temp>=0)
	    temp = FXINT32_MIN;
	else
	    temp = FXINT32_MAX;
    }
    *res = temp;
    return(sat);
}

FX_INLINED fxbool
satsub_i32i32_i32s(fxint32 a, fxint32 b, fxint32* res)
{
    fxint32 temp;
    fxbool sat = FXFALSE;

    temp = a - b;

    if (isoversub_i32i32i32_b(temp, a, b))
    {
	sat = FXTRUE;
	if (temp>=0)
	    temp = FXINT32_MIN;
	else
	    temp = FXINT32_MAX;
    }
    *res = temp;
    return(sat);
}

FX_INLINED fxint16
srx_i16i_i16(fxint16 val, int rshift)
{
    fxint16 res;		/* Result */
    fxint16 rem;		/* Bits shifted off right */

    assert(((unsigned) rshift) < 16);
    if (rshift==0)
	res = val;
    else
    {
	res = val >> rshift;
	rem = val << (16-rshift);
	if (rem<0)
	    res += 1;
	if (rem==-0x8000)
	    res &= 0xfffe;
    }
    return(res);
}

FX_INLINED fxint32
srx_i32i_i32(fxint32 val, int rshift)
{
    fxint32 res;		/* Result */
    fxuint32 rem;		/* Bits shifted off right */

    assert(((unsigned) rshift) < 32);
    if (rshift==0)
	res = val;
    else
    {
	res = sra_i32i_i32(val, rshift);
	rem = val << (32-rshift);
	res += srl_i32i_i32((fxint32) rem, 31); /* If rem has MSB set, add 1 to res */
	if (rem==0x80000000)
	    res &= 0xfffffffe;
    }
    return(res);
}

FX_INLINED fxint64
srx_i64i_i64(fxint64 val, int rshift)
{
    fxint64 res;		/* Result */
    fxint64 rem;		/* Bits shifted off right */

    assert(((unsigned) rshift) < 64);
    if (rshift==0)
	res = val;
    else
    {
	res = sra_i64i_i64(val, rshift);
	rem = sll_i64i_i64(val, 64-rshift);
	if (setlt_i64i64_b(rem, zero_i64()))
	    res += 1;
	if (seteq_i64i64_b(rem, make_i32i32_i64(0x80000000, 0)))
	    res = and_i64i64_i64(res, conv_i32_i64(-2) /* 0xff..ffe */);
    }
    return(res);
}

FX_INLINED fxint32
slx_i32i_i32(fxint32 val, int lshift)
{
    int sign;
    fxuint32 mask;
    fxint32 res;

    assert(((unsigned) lshift) < 32);
    sign = (val < 0);
    mask = 0xffffffff << (32 - lshift - 1);
    if (!sign)
    {
        if (((fxuint32) val & mask) != 0)
            res = FXINT32_MAX;
        else
            res = val << lshift;
    }
    else
    {
        if (((fxuint32) val & mask) != mask)
            res = FXINT32_MIN;
        else
            res = val << lshift;
    }
    return(res);
}

FX_INLINED fxint16
mul_x16x16_x16(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
               const fxint16 x, const fxint16 y)
{
    /* Compensate for multiply moving binary point */
    fxexp_t shift = r_exp - (x_exp+y_exp) + 15;
    fxint16 res;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
        res = fxmul_i16i16_i16(shift, x, y);
    else if (shift>-15 && shift<0)
    {
        fxexcep_expunimp();
        res = 0;
    }
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
        res = 0;
    }
    return(res);
}

FX_INLINED fxint16
add_x16x16_x16(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
               fxint16 x, fxint16 y)
{
    fxint16 a;
    fxint16 b;
    fxint16 res;
    fxexp_t a_b_exp, r_b_exp;
    fxint32 temp;

    if (x_exp >= y_exp)
    {
        a = x;
        b = y;
        a_b_exp = x_exp - y_exp;
        r_b_exp = r_exp - y_exp;
    }
    else
    {
        a = y;
        b = x;
        a_b_exp = y_exp - x_exp;
        r_b_exp = r_exp - x_exp;
    }
    /* Note: saturation might not work for a_b_exp==16 */
    /* Use '&' cos it reduces number of branches */
    if ((a_b_exp<=15) & (((unsigned) r_b_exp)<=(unsigned)(a_b_exp+16)))
    {
        temp = (a<<a_b_exp) + b;
        res = sat_i32_i16(srx_i32i_i32(temp, r_b_exp));
    }
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
        res = 0;
    }
    return(res);
}

FX_INLINED fxint16
sub_x16x16_x16(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
               fxint16 x, fxint16 y)
{
    fxint16 res;
    fxexp_t x_y_exp, r_y_exp;
    fxint32 temp;

    /* Currently only works for x_exp>=y_exp */

    if (x_exp >= y_exp)
    {
	x_y_exp = x_exp - y_exp;
	r_y_exp = r_exp - y_exp;
    }
    else
    {
        fxexcep_expunimp();
        res = 0;
    }

    /* Note: saturation might not work for x_y_exp==16 */
    /* Use '&' cos it reduces number of branches */
    if ((((unsigned) x_y_exp)<=15)
	& (((unsigned) r_y_exp)<=(unsigned) (x_y_exp+16)))
    {
        temp = (x<<x_y_exp) - y;
        res = sat_i32_i16(srx_i32i_i32(temp, r_y_exp));
    }
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
        res = 0;
    }
    return(res);
}

FX_INLINED fxint32
mul_x16x16_x32(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
               const fxint16 x, const fxint16 y)
{
    /* Compensate for multiply moving binary point and extension to 16 bits*/
    fxexp_t shift = r_exp - (x_exp+y_exp) - 1;
    fxint32 res;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
        res = fxmul_i16i16_i32(shift, x, y);
    else if (shift==-1)
    {
	fxint32 temp;

	temp = x * y;
	res = satadd_i32i32_i32(temp, temp);
    }
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
        res = 0;
    }
    return(res);
}

FX_INLINED fxint32
add_x32x32_x32(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
               fxint32 x, fxint32 y)
{
    fxexp_t x_y_exp, shift;
    fxint32 a, b;
    fxint32 res;

/* Currently only handle the case when the result exponent is the same
   as one of the input exponents and the other exponent is smaller
*/

    x_y_exp = x_exp - y_exp;
    if (x_exp==r_exp)
    {
	a = x;
	b = y;
	shift = x_y_exp;
    }
    else if (y_exp==r_exp)
    {
	a = y;
	b = x;
	shift = -x_y_exp;
    }
    else
    {
	a = 0;			/* To stop compiler complaining */
	b = 0;
	shift = -1;		/* Not implemented indicated by neg. shift */
    }

    if (shift == 0)
	res = satadd_i32i32_i32(a, b);
    else if ((unsigned) shift <=31)
	res = _add_x32x32_x32__ls(shift, a, b);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
	res = 0;
    }
    return(res);
}

FX_INLINED fxint32
sub_x32x32_x32(fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
               fxint32 x, fxint32 y)
{
    fxexp_t shift;
    fxint32 res;

/* Currently only handle the case when the result exponent is the same
   as the x input exponent and the other exponent is smaller (but not
   too small!)
*/

    if (x_exp!=r_exp)
	fxexcep_expunimp();
    shift = x_exp - y_exp;

    if (shift == 0)
	res = satsub_i32i32_i32(x, y);
    else if ((unsigned) shift <=31)
	res = _sub_x32x32_x32__ls(shift, x, y);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
	res = 0;
    }
    return(res);
}

FX_INLINED void
add_vx32vx32_vx32(fxsize_t len, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
		  const fxint32 *x, const fxint32 *y, fxint32* res)
{
    fxexp_t x_y_exp, shift;
    const fxint32 *a, *b;

/* Currently only handle the case when the result exponent is the same
   as one of the input exponents and the other exponent is smaller
*/

    x_y_exp = x_exp - y_exp;
    if (x_exp==r_exp)
    {
	a = x;
	b = y;
	shift = x_y_exp;
    }
    else if (y_exp==r_exp)
    {
	a = y;
	b = x;
	shift = -x_y_exp;
    }
    else
    {
	a = NULL;		/* To stop compiler complaining */
	b = NULL;
	shift = -1;		/* Not implemented indicated by neg. shift */
    }

    if (shift == 0)
	satadd_vi32vi32_vi32(len, a, b, res);
    else if ((unsigned) shift <=31)
	_add_vx32vx32_vx32__ls(len, shift, a, b, res);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
acc_vx32_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
	      const fxint32* in, fxint32* res)
{
    add_vx32vx32_vx32(len, in_exp, res_exp, res_exp, in, res, res);
}

FX_INLINED void
add_vx16x16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
		 const fxint16* x, fxint16 y, fxint16* res)
{
    fxexp_t y_x_exp, r_x_exp;

    /* Currently only works for y_exp>=x_exp */

    if (y_exp >= x_exp)
    {
	y_x_exp = y_exp - x_exp;
	r_x_exp = r_exp - x_exp;
    }
    else
    {
	fxexcep_expunimp();
    }

    /* Note: saturation might not work for y_x_exp==16 */
    /* Use '&' cos it reduces number of branches */
    if ((((unsigned) y_x_exp)<=15)
	& (((unsigned) r_x_exp)<=(unsigned) (y_x_exp+16)))
        _add_x16vx16_vx16__lr(size, y_x_exp, r_x_exp, y, x, res);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
sub_x16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                  const fxint16 x, const fxint16* y, fxint16* res)
{
    fxexp_t x_y_exp, r_y_exp;

    /* Currently only works for x_exp>=y_exp */

    if (x_exp>=y_exp)
    {
	x_y_exp = x_exp - y_exp;
	r_y_exp = r_exp - y_exp;
    }
    else
        fxexcep_expunimp();

    /* Note: saturation might not work for x_y_exp==16 */
    /* Use '&' cos it reduces number of branches */
    if ((((unsigned) x_y_exp)<=15)
	& (((unsigned) r_y_exp)<=(unsigned) (x_y_exp+16)))
    {
        _sub_x16vx16_vx16__lr(size, x_y_exp, r_y_exp, x, y, res);
    }
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
mul_vx16x16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                  const fxint16* x, fxint16 y, fxint16* r)
{
    /* Compensate for multiply moving binary point */
    fxexp_t shift = r_exp - (x_exp+y_exp) + 15;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
	mul_srx_vi16i16_i_vi16(size, x, y, shift, r);
    else if (shift>-15 && shift<0)
        fxexcep_expunimp();
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED fxbool
mulacc_vx16x16_vx16s(fxsize_t len, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
		     const fxint16* x, fxint16 y, fxint16* r)
{
    fxbool sat;

    /* Compensate for multiply moving binary point */
    fxexp_t shift = r_exp - (x_exp+y_exp) + (16 - 1);

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
	sat = mul_srx_acc_vi16i16_i_vi16s(len, x, y, shift, r);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
	sat = 0;
    }
    return(sat);
}

FX_INLINED void
mul_vx16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                  const fxint16* x, const fxint16* y, fxint16* r)
{
    /* Compensate for multiply moving binary point */
    fxexp_t shift = r_exp - (x_exp+y_exp) + 15;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
        _mul_vx16vx16_vx16__r(size, shift, x, y, r);
    else if (shift>-15 && shift<0)
        fxexcep_expunimp();
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
add_vx16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                  const fxint16* x, const fxint16* y, fxint16* res)
{
    const fxint16* a;
    const fxint16* b;
    fxexp_t a_b_exp, r_b_exp;

    if (x_exp >= y_exp)
    {
        a = x;
        b = y;
        a_b_exp = x_exp - y_exp;
        r_b_exp = r_exp - y_exp;
    }
    else
    {
        a = y;
        b = x;
        a_b_exp = y_exp - x_exp;
        r_b_exp = r_exp - x_exp;
    }
    /* Note: saturation might not work for a_b==16 */
    /* Use '&' cos it reduces number of branches */
    if ((a_b_exp<=15) & (((unsigned) r_b_exp)<=(unsigned) (a_b_exp+16)))
        _add_vx16vx16_vx16__lr(size, a_b_exp, r_b_exp, a, b, res);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
sub_vx16vx16_vx16(fxsize_t size, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
                  const fxint16* x, const fxint16* y, fxint16* res)
{
    fxexp_t x_y_exp, r_y_exp;

    /* Currently only works for x_exp>=y_exp */

    if (x_exp>=y_exp)
    {
	x_y_exp = x_exp - y_exp;
	r_y_exp = r_exp - y_exp;
    }
    else
	fxexcep_expunimp();

    /* Note: saturation might not work for x_y_exp==16 */
    /* Use '&' cos it reduces number of branches */
    if ((((unsigned) x_y_exp)<=15)
	& (((unsigned) r_y_exp)<=(unsigned) (x_y_exp+16)))
    {
        _sub_vx16vx16_vx16__lr(size, x_y_exp, r_y_exp, x, y, res);
    }
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
mul_vx16x16_vx32(fxsize_t len, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
		 const fxint16* x, fxint16 y, fxint32* r)
{
    /* Compensate for multiply moving binary point and extension to 32 bits*/
    fxexp_t shift = r_exp - (x_exp+y_exp) - 1;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
        _mul_x16vx16_vx32__r(len, shift, y, x, r);
    else if (shift==-1)
	_mul_x16vx16_vx32__l1(len, y, x, r);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
mul_vx16vx16_vx32(fxsize_t len, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
		  const fxint16* x, const fxint16* y, fxint32* r)
{
    /* Compensate for multiply moving binary point and extension to 32 bits*/
    fxexp_t shift = r_exp - (x_exp+y_exp) - 1;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
        _mul_vx16vx16_vx32__r(len, shift, x, y, r);
    else if (shift==-1)
	_mul_vx16vx16_vx32__l1(len, x, y, r);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED void
sub_vx32vx32_vx32(fxsize_t len, fxexp_t x_exp, fxexp_t y_exp, fxexp_t r_exp,
		  const fxint32 *x, const fxint32 *y, fxint32* res)
{
    fxexp_t shift;

/* Currently only handle the case when the result exponent is the same
   as the x input exponent and the other exponent is smaller (but not
   too small!)
*/

    shift = x_exp - y_exp;
    if (shift == 0 && x_exp==r_exp)
	satsub_vi32vi32_vi32(len, x, y, res);
    else if ((unsigned) shift <=31 && x_exp==r_exp)
	_sub_vx32vx32_vx32__ls(len, shift, x, y, res);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
    }
}

FX_INLINED fxbool
mul_vx16mx16_vx32t(fxsize_t height, fxsize_t width, fxexp_t vec_exp,
		   fxexp_t mat_exp, fxexp_t res_exp,
		   const fxint16* vec, const fxint16* mat, fxint32* res)
{
    fxexp_t shift;
    fxbool sat;

    if (width==0 || height==0)
	return(FXFALSE);
    /* Compensate for multiply moving binary point and extension to 32 bits*/
    shift = res_exp - (vec_exp+mat_exp) - 1;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
        sat = mul_srx_vi16mi16_i_vi32t(height, width, vec, mat, shift, res);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
	sat = 0;		/* To suppress compiler warnings */
    }
    return(sat);
}

FX_INLINED fxbool
mul_mx16vx16_vx32t(fxsize_t height, fxsize_t width, fxexp_t mat_exp,
		   fxexp_t vec_exp, fxexp_t res_exp,
		   const fxint16* mat, const fxint16* vec, fxint32* res)
{
    fxexp_t shift;
    fxbool sat;

    if (width==0 || height==0)
	return(FXFALSE);
    /* Compensate for multiply moving binary point and extension to 32 bits*/
    shift = res_exp - (vec_exp+mat_exp) - 1;

    /* Neat trick follows - is 0 <= shift < 32? */
    if (((unsigned) shift) < 32)
        sat = mul_srx_mi16vi16_i_vi32t(height, width, mat, vec, shift, res);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
	sat = 0;		/* To suppress compiler warnings */
    }
    return(sat);
}

FX_INLINED fxbool
mulacc_vx16vx16_mx16s(fxsize_t height, fxsize_t width,
		      fxexp_t a_exp, fxexp_t b_exp, fxexp_t res_exp,
		      const fxint16* a, const fxint16* b, fxint16* res)
{
    fxexp_t rshift;
    fxbool sat;

    if (width==0 || height==0)
	return(FXFALSE);
    /* Compensate for multiply moving binary point */
    rshift = res_exp - (a_exp+b_exp) + (16-1);

    /* Neat trick follows - is 0 <= rshift < 32? */
    if (((unsigned) rshift) < 32)
        sat = mul_srx_acc_vi16vi16_i_mi16s(height, width, a, b, rshift, res);
    else
    {
        fxexcep_exprange();
        fxexcep_expunimp();
	sat = 0;		/* To suppress compiler warnings */
    }
    return(sat);
}

FX_INLINED fxint16
conv_x16_x16(fxexp_t in_exp, fxexp_t res_exp, fxint16 in)
{
    fxint16 res;
    fxexp_t rshift;

    rshift = res_exp - in_exp;

    if (((unsigned) rshift) < 16)
    {
	res = srx_i16i_i16(in, rshift);
    }
    else if (((unsigned) -rshift) < 16)
    {
        res  = satshl_i16i_i16(in, -rshift);
    }
    else if (rshift >= 16)
    {
	/* Right shift>=16 means result always 0 */
	res = 0;
	fxexcep_exprange();
    }
    else if (rshift <= -16)
    {
	/* Left shift >= 16 means result always -FXIN16_MAX, 0 or
	   +XFINT16_MAX */
	res = peg_i16_i16(in);
        fxexcep_exprange();
    }
    return(res);
}

FX_INLINED fxint32
conv_x16_x32(fxexp_t in_exp, fxexp_t res_exp, fxint16 in)
{
    fxint32 res;
    fxexp_t rshift;

    rshift = res_exp - in_exp - (32-16);

    if (((unsigned) rshift) < 16)
    {
        res = srx_i32i_i32(in, rshift);
    }
    else if (((unsigned) -rshift) < 32)
    {
        res  = satshl_i32i_i32(in, -rshift);
    }
    else if (rshift >= 16)
    {
	/* Right shift>=16 means result always 0 */
	res = 0;
	fxexcep_exprange();
    }
    else if (rshift <= -32)
    {
	/* Left shift >= 32 means result always -FXIN32_MAX, 0 or
	   +XFINT32_MAX */
	res = peg_i32_i32(in);
        fxexcep_exprange();
    }
    else
    {
	res = 0;		/* To suppress warning */
	assert(0);
    }
    return(res);
}

FX_INLINED fxint16
conv_x32_x16(fxexp_t in_exp, fxexp_t res_exp, fxint32 in)
{
    fxint16 res;
    fxexp_t rshift;

    rshift = res_exp - in_exp + (32 - 16);

    if (((unsigned) rshift) < 32)
    {
        res = sat_i32_i16(srx_i32i_i32(in, rshift));
    }
    else if (((unsigned) -rshift) < 16)
    {
        res  = satshl_i32i_i16(in, -rshift);
    }
    else if (rshift >=32)
    {
        fxexcep_exprange();
        res = 0;
    }
    else if (rshift <= -16)
    {
	fxexcep_exprange();
	res = peg_i32_i16(in);
    }
    else
    {
	res = 0;		/* To suppress warning */
	assert(0);
    }
    return(res);
}

FX_INLINED fxint32
conv_x32_x32(fxexp_t in_exp, fxexp_t res_exp, fxint32 in)
{
    fxint32 res;
    fxexp_t rshift;

    rshift = res_exp - in_exp;

    if (((unsigned) rshift) < 32)
    {
        res = srx_i32i_i32(in, rshift);
    }
    else if (((unsigned) -rshift) < 32)
    {
        res  = satshl_i32i_i32(in, -rshift);
    }
    else if (rshift >= 32)
    {
	/* Right shift>=16 means result always 0 */
	res = 0;
	fxexcep_exprange();
    }
    else if (rshift <= -32)
    {
	/* Left shift >= 32 means result always -FXIN32_MAX, 0 or
	   +XFINT32_MAX */
	res = peg_i32_i32(in);
        fxexcep_exprange();
    }
    else
    {
	res = 0;		/* To supress warning */
	assert(0);
    }
    return(res);
}

FX_INLINED void
conv_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
	       const fxint16* in, fxint16* res)
{
    fxexp_t rshift;

    rshift = res_exp - in_exp;

    if (((unsigned) rshift) < 16)
    {
        srx_vi16i_vi16(len, in, rshift, res);
    }
    else if (((unsigned) -rshift) < 16)
    {
        _conv_vx16_vx16__l(len, -rshift, in, res);
    }
    else if (rshift >= 16)
    {
	/* Right shift>=16 means result always 0 */
        copy_i16_vi16(len, 0, res);
	fxexcep_exprange();
    }
    else if (rshift <= -16)
    {
	/* Left shift >= 16 means result always -FXINT16_MAX, 0 or
	   +FXINT16_MAX */
	peg_vi16_vi16(len, in, res);
        fxexcep_exprange();
    }
    else
	assert(0);
}

FX_INLINED void
conv_vx32_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
               const fxint32* in, fxint16* res)
{
    fxexp_t rshift;

    rshift = res_exp - in_exp + (32 - 16);

    if (((unsigned) rshift) < 32)
    {
        _conv_vx32_vx16__r(len, rshift, in, res);
    }
    else if (((unsigned) -rshift) < 16)
    {
        _conv_vx32_vx16__l(len, -rshift, in, res);
    }
    else if (rshift <= -16)
    {
        fxexcep_exprange();
	_conv_vx32_vx16__l(len, 16, in, res);
    }
    else if (rshift >= 32)
    {
	fxexcep_exprange();
	copy_i16_vi16(len, 0, res);
    }
    else
	assert(0);
}

FX_INLINED void
conv_vx32_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
	       const fxint32* in, fxint32* res)
{
    fxexp_t rshift;

    rshift = res_exp - in_exp;

    if (((unsigned) rshift) < 32)
    {
        _conv_vx32_vx32__r(len, rshift, in, res);
    }
    else if (((unsigned) -rshift) < 32)
    {
        _conv_vx32_vx32__l(len, -rshift, in, res);
    }
    else if (rshift >= 32)
    {
	/* Right shift>=32 means result always 0 */
        copy_i32_vi32(len, 0, res);
	fxexcep_exprange();
    }
    else if (rshift <= -32)
    {
	/* Left shift >= 32 means result always -FXIN32_MAX, 0 or
	   +XFINT32_MAX */
	peg_vi32_vi32(len, in, res);
        fxexcep_exprange();
    }
    else
	assert(0);
}

FX_INLINED void
conv_vx16_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
	       const fxint16* in, fxint32* res)
{
    fxexp_t rshift;

    rshift = res_exp - in_exp + (16 - 32);

    if (((unsigned) rshift) < 16)
    {
	srx_vi16i_vi32(len, in, rshift, res);
    }
    else if (((unsigned) -rshift) < 32)
    {
	slx_vi16i_vi32(len, in, -rshift, res);
    }
    else if (rshift >= 16)
    {
	/* Right shift>=16 means result always 0 */
        copy_i32_vi32(len, 0, res);
	fxexcep_exprange();
    }
    else if (-rshift >= 32)
    {
	/* Left shift >= 32 means result always -FXINT32_MAX, 0 or
	   +FXINT32_MAX */
	peg_vi16_vi32(len, in, res);
        fxexcep_exprange();
    }
    else
	assert(0);
}

FX_INLINED fxint16
conv_f_x16(fxexp_t res_exp, float val)
{
    fxint16 res;

    conv_f_x16s(res_exp, val, &res);
    return(res);
}

FX_INLINED fxint16
conv_d_x16(fxexp_t res_exp, double val)
{
    fxint16 res;

    conv_d_x16s(res_exp, val, &res);
    return(res);
}

FX_INLINED fxint32
conv_f_x32(fxexp_t res_exp, float val)
{
    fxint32 res;

    conv_f_x32s(res_exp, val, &res);
    return(res);
}

FX_INLINED double
mulpow2_di_d(double in, int n)
{
#if defined(FX_USE_SCALBN)
    return(scalbn(in, n));
#else
    return(scalb(in, n));
#endif
}

FX_INLINED fxint16
ceil_d_x16(fxexp_t res_exp, double in)
{
    double f;
    fxint16 res;

    f = ceil(mulpow2_di_d(in, 16 - 1 - res_exp));
    if (f>(double) FXINT16_MAX)
	res = FXINT16_MAX;
    else if (f<(double) FXINT16_MIN)
	res = FXINT16_MIN;
    else
	res = (fxint16) f;
    return(res);
}

FX_INLINED fxint16
floor_d_x16(fxexp_t res_exp, double in)
{
    double f;
    fxint16 res;

    f = floor(mulpow2_di_d(in, 16 - 1 - res_exp));
    if (f>(double) FXINT16_MAX)
	res = FXINT16_MAX;
    else if (f<(double) FXINT16_MIN)
	res = FXINT16_MIN;
    else
	res = (fxint16) f;
    return(res);
}

FX_INLINED fxint32
ceil_d_x32(fxexp_t res_exp, double in)
{
    fxint32 res;

    res = (fxint32) ceil(mulpow2_di_d(in, 32 - 1 - res_exp));
    return(res);
}

FX_INLINED fxint32
floor_d_x32(fxexp_t res_exp, double in)
{
    fxint32 res;

    res = (fxint32) floor(mulpow2_di_d(in, 32 - 1 - res_exp));
    return(res);
}

FX_INLINED fxbool
lkup_pltable_i16_i16s(const fx_pltable_t* tbl, fxint16 in, fxint16* res)
{
    fxint32 index, offset, base, adjust;
    fxbool sat;

    index = in>>5;
    offset = in & 0x1F;
    base = ( ( (fxint32) tbl[ index + 1024 ] )>>13 )<<6;
    adjust = ( (fxint32) ( tbl[ index + 1024 ]<<19 ) )>>17;
    sat =  sat_i32_i16s( (base + adjust * offset + 0x80) >> 8, res);
    return(sat);
}

FX_INLINED fxint16
lkup_pltable_i16_i16(const fx_pltable_t* tbl, fxint16 val)
{
    fxint32 index, offset, base, adjust;
    fxint32 tabval, res;

    index = val>>5;
    offset = val & 0x1F;
    tabval = tbl[index + 1024];
    base = ( ( (fxint32) tabval)>>13 )<<6;
    adjust = ( (fxint32) ( tabval<<19 ) )>>17;
    res =  sat_i32_i16( (base + adjust * offset + 0x80) >> 8);
    return(res);
}

FX_INLINED fxint16
lkup_gtable_x16_x16(const fx_gtable_t* tbl, fxexp_t in_exp, fxexp_t res_exp,
		    fxint16 val)
{
    fxint16 res;

    if (in_exp==tbl->domain_exp && res_exp==tbl->range_exp)
	res = lkup_pltable_i16_i16(tbl->data, val);
    else
    {
	fxexcep_exprange();
	res = 0;		/* To stop compiler moaning */
    }
    return(res);
}

FX_INLINED void
lkup_gtable_vx16_vx16(fxsize_t len, const fx_gtable_t* tbl,
		      fxexp_t in_exp, fxexp_t res_exp,
		      const fxint16* in, fxint16* res)
{
    if (in_exp==tbl->domain_exp && res_exp==tbl->range_exp)
	lkup_pltable_vi16_vi16(len, tbl->data, in, res);
    else
	fxexcep_exprange();
}

FX_INLINED fxint16
_lkup_utable16_x16_x16__rr(fxint16* base, fxexp_t in_rshift,
			   fxexp_t out_rshift, fxint16 val)
{
    fxint32 temp;
    fxint16 res;

    temp = val >> in_rshift;
    res = srx_i16i_i16(base[temp], out_rshift);
    return(res);
}

FX_INLINED fxint16
lkup_utable16_x16_x16(fx_utable16_t* table, fxexp_t in_exp,
		      fxexp_t res_exp, fxint16 in)
{
    fxexp_t in_rshift;
    fxexp_t out_rshift;
    fxint16 res;

    in_rshift =  table->in_rshift_bias - in_exp;
    out_rshift = res_exp - table->range_exp;
    if ( ((unsigned) in_rshift < 16)
	&& ((unsigned) out_rshift < 16) )
    {
	res = _lkup_utable16_x16_x16__rr(table->base, in_rshift,
					 out_rshift, in);
    }
    else
    {
	fxexcep_exprange();
	fxexcep_expunimp();
	res = 0;
    }
    return(res);
}

FX_INLINED void
lkup_utable16_vx16_vx16(fxsize_t len, fx_utable16_t* table, fxexp_t in_exp,
		      fxexp_t res_exp, fxint16* in, fxint16* res)
{
    fxexp_t in_rshift;
    fxexp_t out_rshift;

    if (in_exp <= table->domain_exp)
    {
	in_rshift =  table->in_rshift_bias - in_exp;
	out_rshift = res_exp - table->range_exp;
	if ( ((unsigned) in_rshift < 16)
	    && ((unsigned) out_rshift == 0) )
	{
	    _lkup_utable16_vx16_vx16__rz(len, table->base, in_rshift, 
					 in, res);
	}
	else
	{
	    fxexcep_exprange();
	    fxexcep_expunimp();
	}
    }
    else
    {
	fxexcep_exprange();
	fxexcep_expunimp();
    }
}

FX_INLINED void
sigmoid_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		  const fxint16* in, fxint16* res)
{
    if (in_exp==4 && res_exp==0)
    {
	lkup_pltable_vi16_vi16(len, pltable_sigmoid_4_0, in, res);
    }
    else
    {
	fxexcep_expunimp();
    }
}

FX_INLINED void
tanh_vx16_vx16(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		  const fxint16* in, fxint16* res)
{
    if (in_exp==3 && res_exp==0)
    {
	lkup_pltable_vi16_vi16(len, pltable_tanh_3_0, in, res);
    }
    else
    {
	fxexcep_expunimp();
    }
}

FX_INLINED void
softmax_vx32_vx32(fxsize_t len, fxexp_t in_exp, fxexp_t res_exp,
		  const fxint32* in, fxint32* res)
{
    fxint32 logsum;

    if (in_exp!=13)
    {
        fxexcep_expunimp();
    }
    else
    {
	logsum = logsum13_vi32_i32(len, in);
	satsub_vi32i32_vi32(len, in, logsum, res);
	exp13_vi32_vx32s(len, res_exp, res, res);
    }
}

#undef FX_INLINED
#ifdef _FX_INLINED
#  define FX_INLINED _FX_INLINED
#  undef _FX_INLINED
#endif

#endif /* #ifndef FXLIB_H_INCLUDED */
